# ---------------------------------------------------------------------
# Gufo Ping: _fast stub
# This is a stub file for pyo3 module only to satisfy
# the mkdocstings
# ---------------------------------------------------------------------
# Copyright (C) 2022-23, Gufo Labs
# ---------------------------------------------------------------------

# Python modules
from typing import Dict, List, Optional

class SocketWrapper(object):
    def __init__(self: "SocketWrapper", afi: int) -> None: ...
    def set_timeout(self: "SocketWrapper", timeout: int) -> None: ...
    def bind(self: "SocketWrapper", addr: str) -> None: ...
    def set_ttl(self: "SocketWrapper", ttl: int) -> None: ...
    def set_unicast_hops(self: "SocketWrapper", ttl: int) -> None: ...
    def set_tos(self: "SocketWrapper", tos: int) -> None: ...
    def set_tclass(self: "SocketWrapper", tos: int) -> None: ...
    def set_send_buffer_size(self: "SocketWrapper", size: int) -> None: ...
    def set_recv_buffer_size(self: "SocketWrapper", size: int) -> None: ...
    def set_coarse(self: "SocketWrapper", coarse: bool) -> None: ...
    def set_accelerated(self: "SocketWrapper", acc: bool) -> None: ...
    def get_fd(self: "SocketWrapper") -> int: ...
    def send(
        self: "SocketWrapper", addr: str, request_id: int, seq: int, size: int
    ) -> None: ...
    def recv(self: "SocketWrapper") -> Optional[Dict[str, int]]: ...
    def get_expired(self: "SocketWrapper") -> Optional[List[str]]: ...
